/*
This is a part of the XProp LiteStep module source code.

Copyright (C) 2000-2003 MrJukes, l0d, repugnant, ilmcuts

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Includes
//
#include "xprop.h"
#include "exports.h"

#pragma comment(lib, "../xPaintClass/Release/xPaintClass-1.0.lib")

#include "../xPaintClass/xExports.h"
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
// Globals
//
int magnify=2;
int ScreenX, ScreenY;
HWND hwndMain, hwndDesktop, hwndParent;
HBITMAP hbmpBG, hCrosshair;
int iCrosshairH, iCrosshairW, iCrosshairX, iCrosshairY;
BOOL crosshair_fInvert;
BITMAP bmpBG;
HPEN pen;
POINT p, o;
COLORREF crosshair_color, color_inverted, clrToInvert;
char format[256] = "";
int timer, offset_x, offset_y;
int x, y;
int mag_x, mag_y, mag_w, mag_h;
int text_x, text_y, text_w, text_h;
int crosshair_size;
BOOL FOLLOW=TRUE, LOCKED=FALSE, FROZEN=FALSE;

PaintText *painttext = NULL;

// Clipboard copying stuff
HBITMAP hbmpClipboard = NULL;
BOOL fEnableClipboard = FALSE;
COLORREF g_clrLastPixel = RGB(0,0,0);

// Color box options
BOOL fEnableColorBox = FALSE;
RECT rColorBox;

// Other stuff
BOOL fOnTop = TRUE, bHooked = FALSE;
BOOL fNotMoveable = FALSE;
BOOL fEnableTrans = FALSE;

// Drawing-only options
int main_h, main_w;

void SetOnTop(BOOL bOnTop)
{
    if (!bHooked && (bOnTop != fOnTop))
    {
        SetWindowPos(hwndMain, bOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
            0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
        
        fOnTop = bOnTop;
    }
}


#ifdef PURELS_FIX

void MyFrame3D(HDC dc, RECT rect, COLORREF TopColor, COLORREF BottomColor, int Width)
{
	HPEN hPen1 = CreatePen(PS_SOLID, 1, TopColor);
	HPEN hPen2 = CreatePen(PS_SOLID, 1, BottomColor);
	HPEN OldPen;
	POINT points[3];

	rect.bottom--;
	rect.right--;

	OldPen = (HPEN)SelectObject(dc, hPen1);

	while (Width > 0)
	{
		POINT p;
		Width--;

		points[0].x = rect.left;
		points[0].y = rect.bottom;
		points[1].x = rect.left;
		points[1].y = rect.top;
		points[2].x = rect.right;
		points[2].y = rect.top;

		Polyline(dc, points, 3);

		SelectObject(dc, hPen2);

		p = points[0];
		points[0] = points[2];
		points[1].x = rect.right;
		points[1].y = rect.bottom;
		points[2] = p;
		points[2].x--;

		Polyline(dc, points, 3);

		SelectObject(dc, hPen1);

		InflateRect(&rect, -1, -1);
	}
	SelectObject(dc, OldPen);
	DeleteObject(hPen1);
	DeleteObject(hPen2);
}
#endif

#ifdef XPROP_RULER

// Ruler stuff
POINT g_StartPoint = {0,0};
POINT g_EndPoint = {0,0};
POINT g_OldPoint = {0,0}; 
BOOL g_fRulerMode = false;

void DrawRuler(const POINT & start, const POINT & end, int spacing, int markSize = 5)
{
	HDC ddc = GetDC(NULL);
	int oldrop = SetROP2(ddc, SRCINVERT);
	MoveToEx(ddc, start.x, start.y, NULL);
	LineTo(ddc, end.x, end.y);


	int dx = end.x - start.x;
	if ( dx < 0 ) dx = dx * -1;
	int dy = end.y - start.y;
	if ( dy < 0 ) dy = dy * -1;

	if ( dx > dy )
	{
		int d = (start.y - end.y) / spacing;
		int y = d + start.y;
		for ( int i = start.x; i < end.x ; i+=spacing )
		{
			MoveToEx(ddc, i, y-markSize,NULL);
			LineTo(ddc, i, y+markSize);
			if ( end.y < start.y )
				y -= d;
			else
				y += d;
		}
	}
	else
	{
		int d = (start.x - end.x) / 10;
		int x = d + start.x;
		for ( int i = start.y; i < end.y ; i+=spacing )
		{
			MoveToEx(ddc,  x-markSize, i, NULL);
			LineTo(ddc,x+markSize, i);
			if ( end.x < start.x )
				x -= d;
			else
				x += d;
		}
			
	}
	
		SetROP2(ddc, oldrop);
		ReleaseDC(NULL, ddc);
}

void BangRuler(HWND caller, const char* args)
{
	if ( g_fRulerMode )
	{
		KillTimer(hwndMain, 1);

		DrawRuler(g_StartPoint, g_OldPoint, 20);
		
		GetCursorPos(&g_EndPoint);
		g_fRulerMode = false;
		char s[255];
		int x = g_EndPoint.x - g_StartPoint.x;
		int y = g_EndPoint.y - g_StartPoint.y;
		if ( x < 0 ) 
			x = x * -1;
		if ( y < 0 )
			y = y * -1;

		sprintf(s, "Delta X: %i\nDelta Y: %i", x, y);
		ReleaseCapture();
		DisplayMessage(s, "Booyah", MB_OK);
	}
	else
	{
		SetCapture(hwndMain);
		SetTimer(hwndDesktop, 1, timer, (TIMERPROC)TimerRuler);
		g_fRulerMode = true;
		GetCursorPos(&g_StartPoint);
		g_OldPoint.x = g_StartPoint.x;
		g_OldPoint.y = g_StartPoint.y;
	}
}

void CALLBACK TimerRuler(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
	if ( g_fRulerMode )
	{
		DrawRuler(g_StartPoint, g_OldPoint, 20);

		GetCursorPos(&g_EndPoint);
		DrawRuler(g_StartPoint, g_EndPoint, 20);
		g_OldPoint.x = g_EndPoint.x;
		g_OldPoint.y = g_EndPoint.y;
		
		
	
	}
}
#endif

void BangToggleInversion(HWND caller, const char* args)
{
	crosshair_fInvert = !crosshair_fInvert;
}

void BangZoomIn(HWND caller, const char* args)
{
	magnify*=2;
	InvalidateRect(hwndMain, NULL, TRUE);
}

void BangZoomOut(HWND caller, const char* args)
{
	magnify/=2;
	if (magnify <= 0) magnify=1;
	InvalidateRect(hwndMain, NULL, TRUE);
}

void BangShow(HWND caller, const char* args) 
{
	SetTimer(hwndMain, 0, timer, (TIMERPROC)TimerMouse);
	ShowWindow(hwndMain, SW_SHOW); 
}

void BangHide(HWND caller, const char* args) 
{ 
	ShowWindow(hwndMain, SW_HIDE);
	KillTimer(hwndMain, 0);
}

void BangToggle(HWND caller, const char* args)
{
	if (IsWindowVisible(hwndMain)) BangHide(caller, args);
	else BangShow(caller, args);
}

void BangSetOrigin(HWND caller, const char* args) 
{ 
	o = p;
	InvalidateRect(hwndMain, NULL, TRUE);
}

void BangResetOrigin(HWND caller, const char* args) 
{ 
	o.x = 0;
	o.y = 0;
	InvalidateRect(hwndMain, NULL, TRUE);
}

void BangLock(HWND caller, const char* args)
{
	if (LOCKED)
	{
		LOCKED=FALSE;
	}
	else
	{
		LOCKED=TRUE;
	}
}

void BangFreeze(HWND caller, const char* args)
{
	if (FROZEN)
	{
		FROZEN=FALSE;
	}
	else
	{
		FROZEN=TRUE;
	}
}

void BangToggleOnTop(HWND caller, const char* args)
{
	SetOnTop(!fOnTop);
}


void BangCopyContents(HWND caller, const char* args)
{
	if (  OpenClipboard(hwndMain) ) {
			EmptyClipboard();
			HBITMAP hTemp = (HBITMAP) CopyImage(hbmpClipboard, IMAGE_BITMAP, 0, 0, 0);
			SetClipboardData(CF_BITMAP, hTemp);
			CloseClipboard();
			DeleteObject(hTemp);
	}
}

void BangCopyHexCode(HWND caller, const char* args)
{
	
	if ( OpenClipboard(hwndMain) ) {
		EmptyClipboard();
		
		//size_t iLen = 7;
		HGLOBAL hGlobal = GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT, 7);
		if ( hGlobal != NULL ) {
			TCHAR * lock = (LPTSTR) GlobalLock(hGlobal);
            sprintf(lock, "%02X%02X%02X", GetRValue(g_clrLastPixel), GetGValue(g_clrLastPixel), GetBValue(g_clrLastPixel));
            //memcpy(lock, s, iLen );
            GlobalUnlock(hGlobal);
            SetClipboardData(CF_TEXT, hGlobal);
        }
    }
    CloseClipboard();
}

void BangMoveable(HWND caller, const char* args)
{
    if (!bHooked)
    {
        fNotMoveable = !fNotMoveable;
    }
}


void BangHook(HWND hCaller, const char* pszArgs)
{
    LPCSTR pszHandle = strrchr(pszArgs,' ');
    
    if (pszHandle)
    {
        HWND hWnd = (HWND)atoi(pszHandle+1);
        if (IsWindow(hWnd))
        {
            SetOnTop(FALSE);
            bHooked = TRUE;
            fNotMoveable = TRUE;
            
            SetWindowLong(hwndMain, GWL_STYLE,
                (GetWindowLong(hwndMain, GWL_STYLE) &~ WS_POPUP)|WS_CHILD);
            SetParent(hwndMain, hWnd);
        }
    }
    return;    
}


void CALLBACK TimerMouse(HWND hWnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
	POINT np;
	GetCursorPos(&np);

	if (p.x != np.x || p.y != np.y)
	{
		p = np;
		if (!LOCKED)
		{
			#ifdef XPROP_RULER
			if ( !g_fRulerMode )
			#endif
				//SetWindowPos(hwndMain, HWND_NOTOPMOST, p.x+offset_x, p.y+offset_y, 0, 0, SWP_NOSIZE | SWP_NOACTIVATE |SWP_NOZORDER);
			//else
				SetWindowPos(hwndMain, HWND_TOPMOST, p.x+offset_x, p.y+offset_y, 0, 0, SWP_NOSIZE | SWP_NOACTIVATE);
		}
	}

	if (!FROZEN) InvalidateRect(hwndMain, NULL, TRUE);



}

void LoadSetup()
{
	char temp[256] = "";

	o.x = 0;
	o.y = 0;

	ScreenX = GetSystemMetrics(SM_CXSCREEN);
	ScreenY = GetSystemMetrics(SM_CYSCREEN);
	hwndDesktop = FindWindow("DesktopBackgroundClass", NULL);
	if (!hwndDesktop) hwndDesktop = GetDesktopWindow();
	if (!hwndDesktop)
	{
		ErrorMessage("Could not locate the desktop window");
		return;
	}

	AddBangCommand("!XPropZoomIn", BangZoomIn);
	AddBangCommand("!XPropZoomOut", BangZoomOut);
	AddBangCommand("!XPropShow", BangShow);
	AddBangCommand("!XPropHide", BangHide);
	AddBangCommand("!XPropToggle", BangToggle);
	AddBangCommand("!XPropSetOrigin", BangSetOrigin);
	AddBangCommand("!XPropResetOrigin", BangResetOrigin);
	AddBangCommand("!XPropFreeze", BangFreeze);
	AddBangCommand("!XPropLock", BangLock);
	AddBangCommand("!XPropToggleInversion", BangToggleInversion);
	AddBangCommand("!XPropToggleOnTop", BangToggleOnTop);
	AddBangCommand("!XPropMoveable", BangMoveable);
	AddBangCommand("!XPropCopyHexCode", BangCopyHexCode);
    AddBangCommand("!XPropLsBoxHook", BangHook);

	#ifdef XPROP_RULER
	AddBangCommand("!XPropRuler", BangRuler);
	#endif
	// ------- Clipboard options
	fEnableClipboard = GetRCBool("XPropEnableClipboard", TRUE);

	if ( fEnableClipboard )
		AddBangCommand("!XPropCopyContents", BangCopyContents);

	// ------- Basic Cross hair options
	crosshair_size = GetRCInt("XPropCrosshairSize", 10);
	crosshair_color = GetRCColor("XPropCrosshairColor", 0x00000000);
	crosshair_fInvert = GetRCBool("XPropCrosshairInvert", TRUE);

	pen = CreatePen(PS_DOT, 0, crosshair_color);
	
	timer = GetRCInt("XPropTimer", 50);
	offset_x = GetRCInt("XPropOffsetX", 25);
	offset_y = GetRCInt("XPropOffsetY", 25);

	x = GetRCCoordinate("XPropX", 0, GetSystemMetrics(SM_CXSCREEN));
	y = GetRCCoordinate("XPropY", 0, GetSystemMetrics(SM_CYSCREEN));
	LOCKED = GetRCBool("XPropStartLocked", TRUE);

	mag_x = GetRCInt("XPropMagWindowX", 6);
	mag_y = GetRCInt("XPropMagWindowY", 6);
	mag_w = GetRCInt("XPropMagWindowWidth", 88);
	mag_h = GetRCInt("XPropMagWindowHeight", 88);

	text_x = GetRCInt("XPropTextWindowX", 7);
	text_y = GetRCInt("XPropTextWindowY", 98);
	text_w = GetRCInt("XPropTextWindowWidth", 88);
	text_h = GetRCInt("XPropTextWindowHeight", 31);

	GetRCString("XPropFormat", format, "X: %d Y: %d\nR: %d G: %d\nB: %d - #%06lx", 256);
	char* c;
	while ((c = strstr(format, "\\n")) != NULL)
	{
		char* q;
		*c='\n';
		c++;
		for (q=c+1; *q; q++, c++)
		{
			*c = *q;
		}
		*c = '\0';
	}
	
	// ------ Color Box Options
	fEnableColorBox = GetRCBool("XPropColorBox", TRUE);
	rColorBox.top = GetRCInt("XPropColorBoxY", 0);
	rColorBox.left = GetRCInt("XPropColorBoxX", 0);
	rColorBox.bottom = GetRCInt("XPropColorBoxHeight", 5) + rColorBox.top;
	rColorBox.right = GetRCInt("XPropColorBoxWidth", 5) + rColorBox.left;

	// ------ "No background bitmap" options
	main_h = GetRCInt("XPropHeight", 135);
	main_w = GetRCInt("XPropWidth", 100);
	
	// we're gonna make our own bitmap, sweet!

	HDC ddc = GetWindowDC(NULL);
	HDC hdcTemp = CreateCompatibleDC(ddc);
	hbmpBG = CreateCompatibleBitmap(ddc, main_w, main_h);
	ReleaseDC(NULL, ddc);
	HGDIOBJ hOld = SelectObject(hdcTemp, hbmpBG);
	
	PaintTexture *paintTexture = Create_xTextureClass();
	paintTexture->configure("xProp", "");
	paintTexture->apply(hdcTemp, 0, 0, main_w, main_h);
	Destroy_xTextureClass(paintTexture);

	int border = GetxRCInt("xProp", "MagWindowBorder", 2, 0);

	PaintTexture *paintTexture1 = Create_xTextureClass();
	paintTexture1->configure("xProp", "MagWindow");
	paintTexture1->apply(hdcTemp, mag_x-border, mag_y-border, mag_w+(2*border), mag_h+(2*border));
	Destroy_xTextureClass(paintTexture1);

	PaintTexture *paintTexture2 = Create_xTextureClass();
	paintTexture2->configure("xProp", "TextWindow");
	paintTexture2->apply(hdcTemp, text_x, text_y, text_w, text_h);
	Destroy_xTextureClass(paintTexture2);

	if ( fEnableColorBox )
	{
		border = GetxRCInt("xProp", "ColorBoxBorder", 0, 0);

		PaintTexture *paintTexture3 = Create_xTextureClass();
		paintTexture3->configure("xProp", "ColorBox");
		paintTexture3->apply(hdcTemp, rColorBox.left-border, rColorBox.top-border, rColorBox.right-rColorBox.left+(2*border), rColorBox.bottom-rColorBox.top+(2*border));
		Destroy_xTextureClass(paintTexture3);
	}

	SelectObject(hdcTemp, hOld);
	if ( !DeleteDC(hdcTemp) )
		ErrorMessage("Deleting temp DC failed");

	// ------ Bitmap cross hair options
	GetRCString("XPropCrosshairBitmap", temp, ".none", 256);
	hCrosshair = LoadLSImage(temp, temp);

	iCrosshairW = GetRCInt("XPropCrosshairBitmapWidth", 15);
	iCrosshairH = GetRCInt("XPropCrosshairBitmapHeight", 15);
	// make a guess for X & Y
	iCrosshairX = GetRCInt("XPropCrosshairBitmapX", (iCrosshairW%2 == 1) ? mag_x+(mag_w/2)-(iCrosshairW/2) : mag_x+(mag_w/2)-(iCrosshairW/2) + 1);
	iCrosshairY = GetRCInt("XPropCrosshairBitmapY", (iCrosshairH%2 == 1) ? mag_y+(mag_h/2)-(iCrosshairH/2) : mag_y+(mag_h/2)-(iCrosshairH/2) + 1);

	color_inverted = GetRCColor("XPropInvertedColor", 0xFF000000);
	if ( color_inverted == 0xFF000000 )
		clrToInvert = RGB(255,255,255);
	else
		clrToInvert = color_inverted;


	// ------ Other settings
	fOnTop = GetRCBool("XPropAlwaysOnTop", TRUE);
	fNotMoveable = GetRCBool("XPropMoveable", FALSE);
	fEnableTrans = GetRCBool("XPropZoomTransparentWindows", TRUE);


}

int initModuleEx(HWND hParent, HINSTANCE hDll, LPCSTR /* pszPath */)
{
	WNDCLASS wc;

	hwndParent = hParent;

	ZeroMemory(&wc,sizeof(wc));
	wc.lpfnWndProc = WndProc;       // our window procedure
	wc.hInstance = hDll;         // hInstance of DLL
	wc.lpszClassName = g_tzModuleName;   // our window class name
	
	if (!RegisterClass(&wc)) 
	{
		ErrorMessage("Could not register window class");
		return 1;
	}

	painttext = Create_xTextClass();
	painttext->configure("xProp", "");
 
	LoadSetup();

	if (hbmpBG)
	{
		int msgs[] = {LM_GETREVID, 0};

		GetObject(hbmpBG, sizeof(bmpBG), &bmpBG);

		hwndMain = CreateWindowEx(WS_EX_TOOLWINDOW, 
								  g_tzModuleName, g_tzModuleName, 
								  WS_POPUP, 
								  x, y, 
								  bmpBG.bmWidth, bmpBG.bmHeight, 
								  NULL, NULL, hDll, NULL);
		SetWindowLong(hwndMain, GWL_USERDATA, MAGIC_DWORD);
		if (!hwndMain) return 1;

		
		SetOnTop(fOnTop);
		SendMessage(hwndParent, LM_REGISTERMESSAGE, (WPARAM)hwndMain, (LPARAM)msgs);

		//SetWindowBitmapRgn(hwndMain, hbmpBG);
		HRGN hrgn = BitmapToRegion(hbmpBG, RGB(255,0,255), 0, 0, 0);
		SetWindowRgn(hwndMain, hrgn, TRUE);

		if (GetRCBool("XPropStartHidden", FALSE))
		{
			SetTimer(hwndMain, 0, timer, (TIMERPROC)TimerMouse);
			ShowWindow(hwndMain, SW_SHOWNORMAL);
		}
	}
	
	return 0;
}

int quitModule(HINSTANCE hDll)
{
	int msgs[] = {LM_GETREVID, 0};

	Destroy_xTextClass(painttext);

	RemoveBangCommand("!XPropZoomIn");
	RemoveBangCommand("!XPropZoomOut");
	RemoveBangCommand("!XPropShow");
	RemoveBangCommand("!XPropHide");
	RemoveBangCommand("!XPropToggle");
	RemoveBangCommand("!XPropSetOrigin");
	RemoveBangCommand("!XPropResetOrigin");
	RemoveBangCommand("!XPropFreeze");
	RemoveBangCommand("!XPropLock");
	RemoveBangCommand("!XPropToggleInversion");
	RemoveBangCommand("!XPropToggleOnTop");
	RemoveBangCommand("!XPropMoveable");
	RemoveBangCommand("!XPropCopyHexCode");
    RemoveBangCommand("!XPropLsBoxHook");
    
    #ifdef XPROP_RULER
    RemoveBangCommand("!XPropRuler");
    #endif

    if ( fEnableClipboard ) {
		RemoveBangCommand("!XPropCopyContents");
		if ( hbmpClipboard )
			DeleteObject(hbmpClipboard);
	}

    DeleteObject(hbmpBG);
	DeleteObject(pen);
	DeleteObject(hCrosshair);
	KillTimer(hwndMain, 0);

	SendMessage(hwndParent, LM_UNREGISTERMESSAGE, (WPARAM)hwndMain, (LPARAM)msgs);
	DestroyWindow(hwndMain);

	UnregisterClass(g_tzModuleName, hDll);

	return 0;
}

// This is a modified version of TransparentBltLS.  I didn't want to use 
// Microsoft's TransparentBlt because on 98 systems it causes memory leaks (and its
// not available on Win95).  This modification handles the inversion.
void TransparentBltXProp( HDC hdcDst, int nXDest, int nYDest, int nWidth, int nHeight, HDC hdcSrc, int nXSrc, int nYSrc, COLORREF colorTransparent)
{
		HDC hdcMem, hdcMask, hdcDstCpy;
		HBITMAP hbmMask, hbmMem, hbmDstCpy;
		HBITMAP hbmOldMem, hbmOldMask, hbmOldDstCpy;

		// create a destination compatble dc containing
		// a copy of the destination dc
		hdcDstCpy	= CreateCompatibleDC(hdcDst);
		hbmDstCpy	= CreateCompatibleBitmap(hdcDst, nWidth, nHeight);
		hbmOldDstCpy = (HBITMAP)SelectObject(hdcDstCpy, hbmDstCpy);

		BitBlt(hdcDstCpy,0,0,nWidth,nHeight,hdcDst,nXDest,nYDest,SRCCOPY);
		
		// create a destination compatble dc containing
		// a copy of the source dc
		hdcMem	= CreateCompatibleDC(hdcDst);
		hbmMem	= CreateCompatibleBitmap(hdcDst, nWidth, nHeight);
		hbmOldMem = (HBITMAP)SelectObject(hdcMem, hbmMem);

		BitBlt(hdcMem,0,0,nWidth,nHeight,hdcSrc,nXSrc,nYSrc,SRCCOPY);

		// the transparent color should be selected as
		// bkcolor into the memory dc
		SetBkColor(hdcMem, colorTransparent);

		// Create monochrome bitmap for the mask
		hdcMask	= CreateCompatibleDC(hdcDst);
		hbmMask = CreateBitmap(nWidth, nHeight, 1, 1, NULL);
		hbmOldMask = (HBITMAP)SelectObject(hdcMask, hbmMask);
		

		// Create the mask from the memory dc
		BitBlt(hdcMask,0,0,nWidth,nHeight,hdcMem,0,0,SRCCOPY);

		// Set the background in hdcMem to black. Using SRCPAINT with black
		// and any other color results in the other color, thus making
		// black the transparent color
		SetBkColor(hdcMem, RGB(0,0,0));
		SetTextColor(hdcMem, RGB(255,255,255));

		BitBlt(hdcMem,0,0,nWidth,nHeight,hdcMask,0,0,SRCAND);

		// Set the foreground to black. See comment above.
		SetBkColor(hdcDst, RGB(255,255,255));
		SetTextColor(hdcDst, RGB(0,0,0));

		BitBlt(hdcDstCpy,0,0,nWidth,nHeight,hdcMask,0,0,SRCAND);

		// Combine the foreground with the background
		BitBlt(hdcDstCpy,0,0,nWidth,nHeight,hdcMem,0,0,SRCPAINT);
		
		if ( crosshair_fInvert ) {
			COLORREF clrOrig, clrInverted;
			int x, y;
			
			// go through each row/column and invert the appropriate pixels
			for ( x = 0; x < nWidth; x++ ) {
				for ( y = 0; y < nHeight; y++ ) {
					clrOrig = GetPixel(hdcDst, x + iCrosshairX, y + iCrosshairY);
					if ( clrOrig > RGB(119,119,119) && clrOrig < RGB(137, 137, 137) )
						clrInverted = RGB(0,0,0);
					else
						clrInverted = RGB( ~GetRValue(clrOrig), ~GetGValue(clrOrig), ~GetBValue(clrOrig));
					
					// if the user hasn't specified an invert color then we invert the whole crosshair
					if ( color_inverted == 0xFF000000  && GetPixel(hdcMask, x, y) != RGB(255,255,255) ) {
						SetPixel(hdcDstCpy, x, y, clrInverted );
					}
					// otherwise just invert the crosshair pixels that are the same as clrToInvert
					else if ( GetPixel(hdcSrc, x, y) == clrToInvert ) {
						SetPixel(hdcDstCpy, x, y, clrInverted );
					}
				}
			}
		}


		// now we have created the image we want to blt
		// in the destination copy dc
		BitBlt(hdcDst,nXDest,nYDest,nWidth,nHeight,hdcDstCpy,0,0,SRCCOPY);

		SelectObject(hdcMask, hbmOldMask);
		DeleteObject(hbmMask);
		DeleteDC(hdcMask);

		SelectObject(hdcMem, hbmOldMem);
		DeleteObject(hbmMem);
		DeleteDC(hdcMem);

		SelectObject(hdcDstCpy, hbmOldDstCpy);
		DeleteObject(hbmDstCpy);
		DeleteDC(hdcDstCpy);
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
        case LM_GETREVID:
        {
            switch (wParam)
            {
                case 0:
                case 1:
                {
                    LPSTR buf = (LPSTR)(lParam);
                    sprintf(buf, "%s (repugnant / V / ilmcuts)", g_tzModuleName);
                    return strlen(buf);
                }

                default:
                    return 0;
            }
        }
        
        case WM_ERASEBKGND: return 0;

		case WM_NCHITTEST:
		{
			if ( ! fNotMoveable ) return HTCAPTION;
		}

		case WM_NCPAINT:
		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hwnd, &ps);
			HDC ddc = GetWindowDC(hwndDesktop);
			HDC buf = CreateCompatibleDC(NULL);
			HDC src = CreateCompatibleDC(NULL);
			HBITMAP bufbmp, oldbuf, oldsrc;
			COLORREF pix;
			char to[256] = "";
		

			bufbmp = CreateCompatibleBitmap(hdc, bmpBG.bmWidth, bmpBG.bmHeight);
			oldbuf = (HBITMAP)SelectObject(buf, bufbmp);

			oldsrc = (HBITMAP)SelectObject(src, hbmpBG);
			BitBlt(buf, 0, 0, bmpBG.bmWidth, bmpBG.bmHeight, src, 0, 0, SRCCOPY);
			SelectObject(src, oldsrc);

			if (!FROZEN) 
			{
				if ( fEnableTrans )
				{
					//BitBlt(buf, mag_x, mag_y, mag_w, mag_h, ddc, p.x-mag_w/(magnify*2), p.y-mag_h/(magnify*2), CAPTUREBLT|SRCCOPY);
					
					HDC transBuf = CreateCompatibleDC(NULL);
					HBITMAP transBmp = CreateCompatibleBitmap(hdc, bmpBG.bmWidth, bmpBG.bmHeight);
					HBITMAP oldTransBmp = (HBITMAP)SelectObject(transBuf, transBmp);
					BitBlt(transBuf, 0, 0, mag_w, mag_h, ddc, p.x-mag_w/(magnify*2), p.y-mag_h/(magnify*2), CAPTUREBLT|SRCCOPY);
					StretchBlt(buf, mag_x, mag_y, mag_w, mag_h, transBuf, 0, 0, mag_w/magnify, mag_h/magnify,SRCCOPY); //CAPTUREBLT
					//StretchBlt(buf, mag_x, mag_y, mag_w, mag_h, transBuf, p.x-mag_w/(magnify*2), p.y-mag_h/(magnify*2), mag_w/magnify, mag_h/magnify, 0x40000000|SRCCOPY); //CAPTUREBLT
					SelectObject(transBuf, oldTransBmp);
					if ( !DeleteObject(transBmp) ) 
						ErrorMessage("Deleting transBmp object failed");
					if ( ! DeleteDC(transBuf) )
						ErrorMessage("Deleting transBuf DC failed");
					
				}
				else {
                    StretchBlt(buf, mag_x, mag_y, mag_w, mag_h, ddc, p.x-mag_w/(magnify*2), p.y-mag_h/(magnify*2), mag_w/magnify, mag_h/magnify, 0x40000000|SRCCOPY); //CAPTUREBLT
				}
				//BitBlt(buf, mag_x, mag_y, mag_w, mag_h, ddc, p.x-mag_w/(magnify*2), p.y-mag_h/(magnify*2), CAPTUREBLT|SRCCOPY);
				if ( fEnableClipboard ) {
					if ( hbmpClipboard) 
						if ( !DeleteObject(hbmpClipboard) )
							ErrorMessage("Unable to delete bitmap.\nPlease email repugnant2k@hotmail.com");
					HDC mdc2 = CreateCompatibleDC(ddc);
					hbmpClipboard = CreateCompatibleBitmap(ddc, mag_w, mag_h);
					HGDIOBJ obj = SelectObject(mdc2, hbmpClipboard);
					//StretchBlt(mdc2, 0, 0, mag_w, mag_h, ddc, p.x-mag_w/(magnify*2), p.y-mag_h/(magnify*2), mag_w/magnify, mag_h/magnify, SRCCOPY);
					BitBlt(mdc2, 0, 0, mag_w, mag_h, buf, mag_x, mag_y, SRCCOPY);
					hbmpClipboard = (HBITMAP) SelectObject(mdc2, obj);
					if ( !DeleteDC(mdc2) ) ErrorMessage("Unable to delete device context.\nPlease email repugnant2k@hotmail.com");
				}
				g_clrLastPixel = pix = GetPixel(buf, mag_x+(mag_w/2), mag_y+(mag_h/2));
				sprintf(to, format, p.x-o.x, p.y-o.y, GetRValue(pix), GetGValue(pix), GetBValue(pix), RGB(GetBValue(pix), GetGValue(pix), GetRValue(pix)));
				//SetRect(&r, text_x, text_y, text_x+text_w, text_y+text_h);

				painttext->apply(buf, text_x, text_y, text_w, text_h, to, true);
				//DrawText(buf, to, strlen(to), &r, DT_EDITCONTROL | DT_WORDBREAK | DT_NOCLIP);

				if ( fEnableColorBox ) {
					HBRUSH hBrush = CreateSolidBrush(pix);
					FillRect(buf, &rColorBox, hBrush);
					DeleteObject(hBrush);
				}
			}

			// if we have a crosshair bitmap then paint that, otherwise use the old style.
			if ( hCrosshair  ) {
				HDC hairBuf = CreateCompatibleDC(NULL);
				SelectObject(hairBuf, hCrosshair);
				TransparentBltXProp(buf, iCrosshairX, iCrosshairY, iCrosshairW, iCrosshairH, hairBuf, 0,0, 0xFF00FF);
				DeleteDC(hairBuf);
			}
			else {
				if ( crosshair_fInvert )
					SetROP2(buf, R2_NOT);

				SelectObject(buf, pen);
				
				MoveToEx(buf, mag_x+(mag_w/2), mag_y+(mag_h/2)-crosshair_size, NULL);
				LineTo(buf, mag_x+(mag_w/2), mag_y+(mag_h/2)+crosshair_size);

				MoveToEx(buf, mag_x+(mag_w/2)-crosshair_size, mag_y+(mag_h/2), NULL);
				LineTo(buf, mag_x+(mag_w/2)+crosshair_size, mag_y+(mag_h/2));

				if ( crosshair_fInvert )
					SetROP2(buf, R2_COPYPEN);
			}

			BitBlt(hdc, 0, 0, bmpBG.bmWidth, bmpBG.bmHeight, buf, 0, 0, SRCCOPY);

			DeleteDC(src);
			DeleteObject(oldsrc);
			SelectObject(buf, oldbuf);
			DeleteObject(bufbmp);
			DeleteDC(buf);
			DeleteObject(oldbuf);
			ReleaseDC(hwndDesktop, ddc);
			EndPaint(hwnd, &ps);
		}
		break;

		case WM_WINDOWPOSCHANGING:
		{
			WINDOWPOS* pos = (WINDOWPOS*)lParam;

			if (pos->x < 0) pos->x = 0;
			else if (pos->x + bmpBG.bmWidth > ScreenX) 
			{
				if (!LOCKED) pos->x = p.x - bmpBG.bmWidth - offset_x;
				else pos->x = ScreenX - bmpBG.bmWidth;
			}

			if (pos->y < 0) pos->y = 0;
			else if (pos->y + bmpBG.bmHeight > ScreenY) 
			{
				if (!LOCKED) pos->y = p.y - bmpBG.bmHeight - offset_y;
				else pos->y = ScreenY - bmpBG.bmHeight;
			}
		}
		break;
	}

	return DefWindowProc(hwnd,msg, wParam,lParam);
}
